from .rest import simple_get


def get_version(nano_handle):
    """gives the version of the api running"""
    # build command (minus the v3 portion)
    version_cmd = nano_handle.url[:-3] + 'version' + '?api-tenant=' + nano_handle.api_tenant
    return simple_get(nano_handle, version_cmd)


def get_buffer_status(nano_handle):
    """ results related to the bytes processed/in the buffer
    """
    status_cmd = nano_handle.url + 'bufferStatus/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    return simple_get(nano_handle, status_cmd)


def get_nano_results(nano_handle, results='All'):
    """ results per pattern
    options:
        ID = cluster ID
        SI = smoothed anomaly index
        RI = raw anomaly index
        FI = frequency index
        DI = distance index
        MD = metadata
    """

    # build results command
    if str(results) == 'All':
        results_str = 'ID,SI,RI,FI,DI'
    else:
        for result in results.split(','):
            if result not in ['ID', 'SI', 'RI', 'FI', 'DI', 'MD']:
                return False, 'unknown result "{}" found in results parameter'.format(result)
        results_str = results

    # build command
    results_cmd = nano_handle.url + 'nanoResults/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    results_cmd += '&results=' + results_str

    return simple_get(nano_handle, results_cmd)


def get_nano_status(nano_handle, results='All'):
    """results in relation to each cluster/overall stats

    results options:
    (includes 0 cluster)
        PCA = principal components
        clusterGrowth = indexes of each increase in cluster
        clusterSizes = number of patterns in each cluster
        anomalyIndexes = anomaly index
        frequencyIndexes = frequency index
        distanceIndexes = distance index

    (overall or no 0 cluster)
        patternMemory = base64 pattern memory
        totalInferences = total number of patterns clustered
        averageInferenceTime = time to cluster per pattern (not available if uploading from serialized nano)
        numClusters = total number of clusters (includes 0 cluster)
    """

    # build results command
    if str(results) == 'All':
        results_str = 'PCA,clusterGrowth,clusterSizes,anomalyIndexes,frequencyIndexes,distanceIndexes,totalInferences,numClusters'
    else:
        for result in results.split(','):
            if result not in ['PCA', 'clusterGrowth', 'clusterSizes', 'anomalyIndexes', 'frequencyIndexes',
                              'distanceIndexes', 'totalInferences', 'numClusters', 'averageInferenceTime']:
                return False, 'unknown result "{}" found in results parameter'.format(result)
        results_str = results

    # build command
    results_cmd = nano_handle.url + 'nanoStatus/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    results_cmd = results_cmd + '&results=' + results_str

    return simple_get(nano_handle, results_cmd)



from urllib3 import PoolManager
from urllib3 import Timeout
import json
from os.path import expanduser
from os import path
from os import environ
import tarfile
from .rest import simple_get
from .rest import simple_delete
from .rest import simple_post


############################
# BoonNano Python API v3 #
############################

class BoonException(Exception):
    def __init__(self, message):
        self.message = message


def http_msg(response):
    return '{}:{}'.format(response.status, response.reason)


class NanoHandle:

    def __init__(self, user, license="~/.BoonLogic", timeout=60.0):

        self.instance = ''
        self.numeric_format = ''

        license_path = expanduser(license)

        if path.exists(license_path):
            try:
                with open(license_path, "r") as json_file:
                    file_data = json.load(json_file)
            except json.JSONDecodeError as e:
                raise BoonException(
                    "json formatting error in .BoonLogic file, {}, line: {}, col: {}".format(e.msg, e.lineno, e.colno))

        # load the user, environment gets precedence
        if 'BOON_USER' in environ:
            self.user = environ['BOON_USER']
            license_block = dict()
        else:
            if user not in file_data:
                raise BoonException("'{}' is missing from configuration, set via BOON_USER or in ~/.BoonLogic".format(user))
            license_block = file_data[user]

        # load the api-key, environment gets precedence
        if 'BOON_API_KEY' in environ:
            self.api_key = environ('BOON_API_KEY')
        else:
            if 'api-key' not in license_block.keys():
                raise BoonException("'api-key' is missing from configuration, set via BOON_API_KEY or in ~/.BoonLogic file")
            self.api_key = license_block['api-key']

        # load the server, environment gets precedence
        if 'BOON_SERVER' in environ:
            self.server = environ("BOON_SERVER")
        else:
            if 'server' not in license_block.keys():
                raise BoonException("'server' is missing from configuration, set via BOON_SERVER or in ~/.BoonLogic file")
            self.server = license_block['server']

        # load the tenant, environment gets precedence
        if 'BOON_TENANT' in environ:
            self.api_tenant = environ("BOON_TENANT")
        else:
            if 'api-tenant' not in license_block.keys():
                raise BoonException(
                    "'api-tenant' is missing from configuration, set via BOON_TENANT or in ~/.BoonLogic file")
            self.api_tenant = license_block['api-tenant']

        # set up base url
        self.url = self.server + '/expert/v3/'
        if "http" not in self.server:
            self.url = "http://" + self.url

        # create pool manager
        timeoutInst = Timeout(connect=30.0, read=timeout)
        self.http = PoolManager(timeout=timeoutInst)


# start the nano and create the unique nano handle
def open_nano(nano_handle, instance_id):
    """Creates or attaches to a nano pod instance

    Args:
        nano_handle (NanoHandle): handle for this nano instance
        instance_id (str): instance identifier to be applied to instance

    Returns:
        success status
        error message

    """
    success, response = create_instance(nano_handle, instance_id)
    if not success:
        return False, response

    return True, None


# free the nano label instance and close the http connection
def close_nano(nano_handle):
    """Closes the pod instance

    Args:
        nano_handle (NanoHandle): handle for this nano instance

    Returns:
        success status
        error message

    """
    close_cmd = nano_handle.url + 'nanoInstance/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant

    # delete instance
    result, response = simple_delete(nano_handle, close_cmd)
    if result:
        nano_handle.http.clear()

    return result, response


# get the labels of running nanos
def nano_list(nano_handle):
    """returns list of nano instances running
    """

    # build command
    instance_cmd = nano_handle.url + 'nanoInstances' + '?api-tenant=' + nano_handle.api_tenant

    return simple_get(nano_handle, instance_cmd)


# store the nano for later use
def save_nano(nano_handle, filename):
    """serializes the nano and saves it as a tar filename
    """

    # build command
    snapshot_cmd = nano_handle.url + 'snapshot/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant

    # serialize nano
    try:
        snapshot_response = nano_handle.http.request(
            'GET',
            snapshot_cmd,
            headers={
                'x-token': nano_handle.api_key
            }
        )
    except Exception as e:
        return False, "request failed"

    # check for error
    if snapshot_response.status != 200:
        return False, http_msg(snapshot_response)

    # at this point, the call succeeded, saves the result to a local file
    try:
        with open(filename, 'wb') as fp:
            fp.write(snapshot_response.data)
    except Exception as e:
        return False, 'unable to write file'

    return True, None


def restore_nano(nano_handle, filename):
    """deserialize existing nano
    upload file to given instance

    NOTE: must be of type tar
    """

    # verify that input file is a valid nano file (gzip'd tar with Magic Number)
    try:
        with tarfile.open(filename, 'r:gz') as tp:
            with tp.extractfile('BoonNano/MagicNumber') as magic_fp:
                magic_num = magic_fp.read()
                if magic_num != b'\xef\xbe':
                    return False, 'file {} is not a Boon Logic nano-formatted file, bad magic number'.format(filename)

    except KeyError as ke:
        return False, 'file {} is not a Boon Logic nano-formatted file'.format(filename)

    with open(filename, 'rb') as fp:
        nano = fp.read()

    # build command
    snapshot_cmd = nano_handle.url + 'snapshot/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant

    # post serialized nano
    try:
        snapshot_response = nano_handle.http.request(
            'POST',
            snapshot_cmd,
            headers={
                'x-token': nano_handle.api_key
            },
            fields={
                'snapshot': (filename, nano)
            }
        )

    except Exception as e:
        return False, 'request failed'

    # check for error
    if snapshot_response.status != 200:
        return False, http_msg(snapshot_response)

    nano_handle.numeric_format = json.loads(snapshot_response.data.decode('utf-8'))['numericFormat']

    return True, None


###########
# PRIVATE #
###########

def create_instance(nano_handle, label):
    # build command
    instance_cmd = nano_handle.url + 'nanoInstance/' + label + '?api-tenant=' + nano_handle.api_tenant

    success, response = simple_post(nano_handle, instance_cmd)
    if success:
        nano_handle.instance = label

    return success, response



from .management import NanoHandle
from .management import BoonException
from .management import open_nano
from .management import close_nano
from .management import nano_list
from .management import save_nano
from .management import restore_nano
from .management import http_msg

from .configure import get_config
from .configure import generate_config
from .configure import configure_nano
from .configure import autotune_config

from .cluster import load_data
from .cluster import load_file
from .cluster import run_nano

from .results import get_version
from .results import get_buffer_status
from .results import get_nano_status
from .results import get_nano_results

setattr(NanoHandle, "open_nano", open_nano)
setattr(NanoHandle, "close_nano", close_nano)
setattr(NanoHandle, "nano_list", nano_list)
setattr(NanoHandle, "save_nano", save_nano)
setattr(NanoHandle, "restore_nano", restore_nano)

setattr(NanoHandle, "get_config", get_config)
setattr(NanoHandle, "configure_nano", configure_nano)
setattr(NanoHandle, "generate_config", generate_config)
setattr(NanoHandle, "autotune_config", autotune_config)

setattr(NanoHandle, "load_data", load_data)
setattr(NanoHandle, "load_file", load_file)
setattr(NanoHandle, "run_nano", run_nano)

setattr(NanoHandle, "get_version", get_version)
setattr(NanoHandle, "get_buffer_status", get_buffer_status)
setattr(NanoHandle, "get_nano_status", get_nano_status)
setattr(NanoHandle, "get_nano_results", get_nano_results)



import os
import numpy as np
from .rest import simple_post
from .rest import multipart_post


def load_file(nano_handle, file, file_type, gzip=False, metadata=None, append_data=False):
    """posts data to the nano
    """

    # load the data file
    try:
        with open(file) as fp:
            file_data = fp.read()
    except FileNotFoundError as e:
        return False, e.strerror

    # verify file_type is set correctly
    if file_type not in ['csv', 'raw']:
        return False, 'file_type must be "csv" or "raw"'

    file_name = os.path.basename(file)

    if metadata:
        fields = {'data': (file_name, file_data),
                  'metadata': metadata.replace(',', '|').replace('{', '').replace('}', '').replace(' ', '')}
    else:
        fields = {'data': (file_name, file_data)}

    # build command
    dataset_cmd = nano_handle.url + 'data/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    dataset_cmd += '&fileType=' + file_type
    dataset_cmd += '&gzip=' + str(gzip).lower()
    dataset_cmd += '&appendData=' + str(append_data).lower()

    return multipart_post(nano_handle, dataset_cmd, fields=fields)


def load_data(nano_handle, data, metadata=None, append_data=False):

    if not isinstance(data, np.ndarray):
        if nano_handle.numeric_format == 'int16':
            data = np.asarray(data, dtype=np.int16)
        elif nano_handle.numeric_format == 'float32':
            data = np.asarray(data, dtype=np.float32)
        elif nano_handle.numeric_format == 'uint16':
            data = np.asarray(data, dtype=np.uint16)

    if nano_handle.numeric_format == 'int16':
        data = data.astype(np.int16)
    elif nano_handle.numeric_format == 'float32':
        data = data.astype(np.float32)
    elif nano_handle.numeric_format == 'uint16':
        data = data.astype(np.uint16)
    data = data.tostring()
    file_name = 'dummy_filename.bin'
    file_type = 'raw'

    if metadata:
        fields = {'data': (file_name, data),
                  'metadata': metadata.replace(',', '|').replace('{', '').replace('}', '').replace(' ', '')}
    else:
        fields = {'data': (file_name, data)}

    # build command
    dataset_cmd = nano_handle.url + 'data/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    dataset_cmd += '&fileType=' + file_type
    dataset_cmd += '&appendData=' + str(append_data).lower()

    return multipart_post(nano_handle, dataset_cmd, fields=fields)


def run_nano(nano_handle, results=None):
    """ clusters the data in the buffer
    returns any specified results

    results per pattern options:
        ID = cluster ID
        SI = smoothed anomaly index
        RI = raw anomaly index
        FI = frequency index
        DI = distance index
        MD = metadata
    """

    results_str = ''
    if str(results) == 'All':
        results_str = 'ID,SI,RI,FI,DI'
    elif results:
        for result in results.split(','):
            if result not in ['ID', 'SI', 'RI', 'FI', 'DI', 'MD']:
                return False, 'unknown result "{}" found in results parameter'.format(result)
        results_str = results

    # build command
    nano_cmd = nano_handle.url + 'nanoRun/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    if results:
        nano_cmd += '&results=' + results_str

    return simple_post(nano_handle, nano_cmd)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='boonnano',
    version='3.0.11',
    author="BoonLogic",
    author_email="elise@boonlogic.com",
    packages=['boonnano'],
    install_requires=['urllib3','numpy'],
    description="A SDK package for utilizing the BoonLogic nano API",
    long_description=long_description,
    license='MIT',
    long_description_content_type="text/markdown",
    url="https://github.com/boonlogic/Python_API",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ]
)



import json
import boonnano
import urllib3


def simple_get(nano_handle, get_cmd):

    try:
        response = nano_handle.http.request(
            'GET',
            get_cmd,
            headers={
                'x-token': nano_handle.api_key,
                'Content-Type': 'application/json'
            }
        )
    except Exception as e:
        return False, 'request failed'

    # check for error
    if response.status != 200:
        return False, boonnano.http_msg(response)

    return True, json.loads(response.data.decode('utf-8'))


def multipart_post(nano_handle, post_cmd, fields=None):

    multi_fields = urllib3.filepost.encode_multipart_formdata(fields)

    try:
        response = nano_handle.http.request(
            'POST',
            post_cmd,
            headers={
                'x-token': nano_handle.api_key
            },
            fields=fields
        )
    except Exception as e:
        return False, 'request failed: {}'.format(e)

    if response.status != 200:
        return False, boonnano.http_msg(response)

    if len(response.data) == 0:
        return True, ''

    return True, json.loads(response.data.decode('utf-8'))


def simple_post(nano_handle, post_cmd, body=None):

    try:
        response = nano_handle.http.request(
            'POST',
            post_cmd,
            headers={
                'x-token': nano_handle.api_key,
                'Content-type': 'application/json'
            },
            body=body
        )
    except Exception as e:
        return False, 'request failed: {}'.format(e)

    # check for error
    if response.status != 200:
        return False, boonnano.http_msg(response)

    if len(response.data) == 0:
        return True, ''

    return True, json.loads(response.data.decode('utf-8'))


def simple_delete(nano_handle, delete_cmd):

    try:
        response = nano_handle.http.request(
            'DELETE',
            delete_cmd,
            headers={
                'x-token': nano_handle.api_key
            }
        )
    except Exception as e:
        return False, 'request failed: {}'.format(e)

    # check for error
    if response.status != 200:
        return False, boonnano.http_msg(response)

    return True, json.loads(response.data.decode('utf-8'))



import json
from .rest import simple_get
from .rest import simple_post


def configure_nano(nano_handle, feature_count=10, numeric_format="float32", min=1, max=10, weight=1, labels="",
                   percent_variation=0.05, streaming_window=1, accuracy=0.99, config=None):
    """returns the posted clustering configuration
    """

    # verify numeric_format
    if numeric_format not in ['float32', 'int16', 'uint16']:
        return False, 'numeric_format must be "float32", "int16", or "uint16"'

    # build command
    config_cmd = nano_handle.url + 'clusterConfig/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    if not config:
        new_config = generate_config(numeric_format, feature_count, min, max, weight, labels, percent_variation,
                                     streaming_window, accuracy)
    else:
        new_config = config

    body = json.dumps(new_config)

    result, response = simple_post(nano_handle, config_cmd, body=body)
    if result:
        nano_handle.numeric_format = new_config['numericFormat']

    return result, response


def generate_config(numeric_format, feature_count, min=1, max=10, weight=1, labels="", percent_variation=0.05,
                    streaming_window=1, accuracy=0.99):
    """returns a json config version for the given Parameters
    """
    config = {}
    config['accuracy'] = accuracy
    temp_array = []
    for x in range(feature_count):
        temp_feature = {}
        # max
        if len([max]) == 1:
            temp_feature['maxVal'] = max
        else:  # the max vals are given as a list
            temp_feature['maxVal'] = max[x]
        # min
        if len([min]) == 1:
            temp_feature['minVal'] = min
        else:  # the min vals are given as a list
            temp_feature['minVal'] = min[x]
        # weights
        if len([weight]) == 1:
            temp_feature['weight'] = weight
        else:  # the weight vals are given as a list
            temp_feature['weight'] = weight[x]
        # labels
        if labels != "" and labels[x] != "":
            temp_feature['label'] = labels[x]
        temp_array.append(temp_feature)
    config['features'] = temp_array
    config['numericFormat'] = str(numeric_format)
    config['percentVariation'] = percent_variation
    config['streamingWindowSize'] = streaming_window

    return config


def autotune_config(nano_handle, autotune_pv=True, autotune_range=True, by_feature=False, exclusions={}):
    """autotunes the percent variation
    and the min and max for each feature
    """

    # build command
    config_cmd = nano_handle.url + 'autoTuneConfig/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    config_cmd += '&byFeature=' + str(by_feature).lower()
    config_cmd += '&autoTunePV=' + str(autotune_pv).lower()
    config_cmd += '&autoTuneRange=' + str(autotune_range).lower()
    config_cmd += '&exclusions=' + str(exclusions)[1:-1].replace(' ', '')

    # autotune parameters
    return simple_post(nano_handle, config_cmd)


def get_config(nano_handle):
    """returns the posted clustering configuration
    """
    config_cmd = nano_handle.url + 'clusterConfig/' + nano_handle.instance + '?api-tenant=' + nano_handle.api_tenant
    return simple_get(nano_handle, config_cmd)
